#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Stéphane Caron.
#
# This file is part of Pink.
#
# Pink is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Pink is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pink. If not, see <http://www.gnu.org/licenses/>.

"""
Python inverse kinematics for your robot model based on Pinocchio.
"""

from .configured_robot import ConfiguredRobot
from .configured_robot import assume_robot_is_configured
from .configured_robot import configure_robot
from .solve_ik import solve_ik

__version__ = "0.0.2"

__all__ = [
    "ConfiguredRobot",
    "assume_robot_is_configured",
    "configure_robot",
    "models",
    "solve_ik",
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Stéphane Caron.
#
# This file is part of Pink.
#
# Pink is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Pink is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pink. If not, see <http://www.gnu.org/licenses/>.

"""
Test transform functions.
"""

import unittest
import warnings

import numpy as np

from pink import get_transform_body_to_world

from .models import build_jvrc_model


class TestTransforms(unittest.TestCase):
    def setUp(self):
        """
        Prepare test fixture.
        """
        warnings.simplefilter("ignore", category=DeprecationWarning)
        warnings.simplefilter("ignore", category=UserWarning)
        self.robot = build_jvrc_model()
        self.assertIsNotNone(self.robot)

    def test_transform_found(self):
        """
        Get the pose of an existing robot body.
        """
        transform_pelvis_to_world = get_transform_body_to_world(
            self.robot, "PELVIS_S"
        )
        self.assertTrue(
            np.allclose(
                transform_pelvis_to_world.np[3, :],
                np.array([0.0, 0.0, 0.0, 1.0]),
            )
        )

    def test_transform_not_found(self):
        """
        Raise an error when the robot body is not found.
        """
        with self.assertRaises(KeyError):
            get_transform_body_to_world(self.robot, "foo")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Stéphane Caron.
#
# This file is part of Pink.
#
# Pink is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Pink is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pink. If not, see <http://www.gnu.org/licenses/>.

"""
Function to solve inverse kinematics.
"""

from typing import Iterable

import numpy as np
import pinocchio as pin

from qpsolvers import solve_qp

from .checks import assert_configuration_is_within_limits
from .limits import compute_velocity_limits
from .tasks import Task


def compute_qp_objective(
    robot: pin.RobotWrapper, tasks: Iterable[Task], damping: float
):
    """
    Compute the Hessian matrix :math:`H` and linear vector :math:`c` of the
    QP objective function:

    .. math::

        \\frac{1}{2} \\Delta q^T H \\Delta q + c^T q

    The configuration displacement :math:`\\Delta q` is the output of inverse
    kinematics (we divide it by :math:`\\Delta t` to get a commanded velocity).

    Args:
        robot: Robot model and its current configuration :math:`q`.
        tasks: List of kinematic tasks to fulfill at (weighted) best.
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            `[cost]^2 / [tangent]` where `[tangent]` is "the" unit of robot
            velocities. Improves numerical stability, but larger values slow
            down all tasks.

    Returns:
        Pair :math:`(H, c)` of Hessian matrix and linear vector of the QP
        objective.
    """
    H = damping * np.eye(robot.nv)
    c = np.zeros((robot.nv,))
    for task in tasks:
        H_task, c_task = task.compute_qp_objective(robot)
        H += H_task
        c += c_task
    return (H, c)


def solve_ik(
    robot: pin.RobotWrapper,
    tasks: Iterable[Task],
    dt: float,
    damping: float = 1e-12,
    solver: str = "quadprog",
) -> np.ndarray:
    """
    Compute a velocity tangent to the current robot configuration that
    satisfies at (weighted) best a given set of kinematic tasks.

    This function assumes that the robot data already underwent forward
    kinematics and frame placement update. TODO(scaron): properly handled
    by ConfiguredRobot type.

    Args:
        robot: Robot model.
        tasks: List of kinematic tasks.
        dt: Integration timestep in [s].
        damping: weight of Tikhonov (everywhere) regularization. Its unit is
            :math:`[\\mathrm{cost}]^2 / [\\mathrm{tangent}]` where
            :math:`[\\mathrm{tangent}]` is "the" unit of robot velocities.
            Improves numerical stability, but larger values slow down all
            tasks.
        solver: Backend quadratic programming solver.

    Returns:
        Velocity :math:`v` in tangent space, of dimension `robot.nv`.

    Raises:
        NotWithinConfigurationLimits: if the current robot configuration is not
            within limits.

    Note:
        Our Tikhonov damping is isotropic despite tangent velocities not being
        homogeneous. If it helps we can add a tangent-space scaling to damp the
        floating base differently from joint angular velocities.
    """
    assert_configuration_is_within_limits(robot)
    H, c = compute_qp_objective(robot, tasks, damping)
    v_max, v_min = compute_velocity_limits(robot, dt)
    tangent_eye = np.eye(robot.nv)
    A = np.vstack([tangent_eye, -tangent_eye])
    b = np.hstack([v_max * dt, -v_min * dt])
    Delta_q = solve_qp(H, c, A, b, solver=solver)
    assert Delta_q is not None
    v: np.ndarray = Delta_q / dt
    return v



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Stéphane Caron.
#
# This file is part of Pink.
#
# Pink is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Pink is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pink. If not, see <http://www.gnu.org/licenses/>.

"""
Test inverse kinematics.
"""

import unittest

import numpy as np

from pink import solve_ik
from pink.models import build_jvrc_model


class TestSolveIK(unittest.TestCase):
    def setUp(self):
        """
        Prepare test fixture.
        """
        self.robot = build_jvrc_model()
        self.assertIsNotNone(self.robot)

    def test_no_task(self):
        """
        Raise an error when the robot body is not found.
        """
        tasks = []
        dt = 1e-3  # [s]
        v = solve_ik(self.robot, tasks, dt)
        self.assertTrue(np.allclose(v, np.zeros(self.robot.nv)))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Stéphane Caron.
#
# This file is part of Pink.
#
# Pink is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Pink is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pink. If not, see <http://www.gnu.org/licenses/>.

"""
Joint limits implemented as inequality constraints.
"""

from typing import Tuple

import numpy as np
import pinocchio as pin


def compute_velocity_limits(
    robot: pin.RobotWrapper,
    dt: float,
    config_limit_gain: float = 0.5,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Compute the robot's configuration-dependent velocity limits:

    .. math::

        v_{min}(q) \\leq v \\leq v_{max}(q)

    where :math:`q \\in {\\cal C}` is the robot's configuration and :math:`v
    \\in T_q({\\cal C})` is the velocity in the tangent space at :math:`q`.
    These velocity limits combine:

    - Configuration-independent joint velocity limits
      :math:`|\\dot{q}| \\leq \\dot{q}_{lim}`.
    - Time-derivatives of the configuration limit
      :math:`q_{min} \\leq q \\leq q_{max}`.

    Args:
        robot: Robot model and configuration.
        dt: Integration timestep in [s].
        config_limit_gain: gain between 0 and 1 to steer away from
            configuration limits. It is described in "Real-time prioritized
            kinematic control under inequality constraints for redundant
            manipulators" (Kanoun, 2012). More details in [this
            writeup](https://scaron.info/teaching/inverse-kinematics.html).

    Returns:
        Pair $(v_{max}(q), v_{min}(q))$ of velocity lower and upper bounds.
    """
    v_max = +1000.0 * np.ones(robot.nv)
    v_min = -1000.0 * np.ones(robot.nv)
    return v_max, v_min



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Stéphane Caron.
#
# This file is part of Pink.
#
# Pink is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Pink is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pink. If not, see <http://www.gnu.org/licenses/>.

"""
Test base class for kinematic tasks.
"""

import unittest

from pink.models import build_jvrc_model
from pink.tasks import Task


class TestTask(unittest.TestCase):
    def setUp(self):
        """
        Prepare test fixture.
        """
        self.robot = build_jvrc_model()
        self.assertIsNotNone(self.robot)

    def test_task_not_implemented(self):
        """
        Raise an error when the robot body is not found.
        """
        task = Task()
        self.assertIsNotNone(task)
        with self.assertRaises(NotImplementedError):
            task.compute_task_dynamics(self.robot)
        with self.assertRaises(NotImplementedError):
            task.compute_qp_objective(self.robot)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Stéphane Caron.
#
# This file is part of Pink.
#
# Pink is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Pink is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pink. If not, see <http://www.gnu.org/licenses/>.

"""
Pink adds a notion of *configured* robot to indicate that forward kinematics
functions have been run, so that frame transforms and frame Jacobians used for
IK can be queried from the robot's data.

TODO(scaron): two versions of solve_ik, with/without configuration step.
"""

import numpy as np
import pinocchio as pin


class ConfiguredRobot:

    """
    Type annotation indicating that quantities that depend on the configuration
    :math:`q` have been computed.

    In Pink, this type enables access to frame transforms and frame Jacobians.
    We rely on typing to make sure the proper forward kinematics functions have
    been called beforehand. In Pinocchio, these functions are:

    - ``pin.computeJointJacobians(model, data, configuration)``
    - ``pin.updateFramePlacements(model, data)``

    The former computes the full model Jacobian into ``data.J``. (It also
    computes forward kinematics, so there is no need to further call
    ``pin.forwardKinematics(model, data, configuration)``.) The latter updates
    frame placements.
    """

    def __init__(self, robot: pin.RobotWrapper):
        self.collision_data = robot.collision_data
        self.collision_model = robot.collision_model
        self.data = robot.data
        self.model = robot.model
        self.q0 = robot.q0
        self.v0 = robot.v0
        self.visual_data = robot.visual_data
        self.visual_model = robot.visual_model
        self.viz = robot.viz

    def get_body_jacobian(self, body_name: str) -> np.ndarray:
        """
        Compute the Jacobian matrix :math:`{}_B J_{WB}` of the body velocity
        :math:`{}_B v_{WB}`:

        .. math::

            {}_B v_{WB} = {}_B J_{WB} \\dot{q}

        Args:
            body_name: Name of robot body.

        Returns:
            Jacobian :math:`{}_B J_{WB}` of the body twist.

        When the robot model includes a floating base
        (pin.JointModelFreeFlyer), the configuration vector :math:`q` consists
        of:

        - ``q[0:3]``: position in [m] of the floating base in the inertial
          frame, formatted as :math:`[p_x, p_y, p_z]`.
        - ``q[3:7]``: unit quaternion for the orientation of the floating base
          in the inertial frame, formatted as :math:`[q_x, q_y, q_z, q_w]`.
        - ``q[7:]``: joint angles in [rad].
        """
        body_id = self.model.getBodyId(body_name)
        J: np.ndarray = pin.getFrameJacobian(
            self.model, self.data, body_id, pin.ReferenceFrame.LOCAL
        )
        return J

    def get_transform_body_to_world(self, body_name: str) -> pin.SE3:
        """
        Get the pose of a body frame of the robot in its current configuration.

        Args:
            body: Body name.

        Returns:
            Current transform from body frame to world frame.

        Raises:
            KeyError: if the body name is not found in the robot model.
        """
        body_id = self.model.getBodyId(body_name)
        try:
            return self.data.oMf[body_id].copy()
        except IndexError as index_error:
            raise KeyError(
                f'Body "{body_name}" not found in robot model'
            ) from index_error


def assume_robot_is_configured(robot: pin.RobotWrapper) -> ConfiguredRobot:
    """
    Assume that the provided robot wrapper has already been configured.

    Returns:
        Same robot represented with a different type.
    """
    return ConfiguredRobot(robot)


def configure_robot(robot: pin.RobotWrapper, q: np.ndarray) -> ConfiguredRobot:
    """
    Run forward kinematics on a robot wrapper.

    Args:
        robot: Robot wrapper with its initial data.
        q: Configuration to apply.

    Returns:
        Configured robot.
    """
    pin.computeJointJacobians(robot.model, robot.data, q)
    pin.updateFramePlacements(robot.model, robot.data)
    return ConfiguredRobot(robot)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Stéphane Caron.
#
# This file is part of Pink.
#
# Pink is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Pink is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Pink. If not, see <http://www.gnu.org/licenses/>.

"""
All kinematic tasks derive from the following base class. The formalism used in
this implementation is written down in `this note on task-based inverse
kinematics <https://scaron.info/robotics/inverse-kinematics.html>`_. As of
February 2022 it hasn't been updated with the proper dimensional analysis, but
the core concepts and notations are there.
"""

from typing import Tuple

import numpy as np
import pinocchio as pin


class Task:

    """
    Base class for kinematic tasks.

    Attributes:
        gain: Task gain :math:`\\alpha \\in [0, 1]` for additional low-pass
            filtering. Defaults to 1.0 (no filtering) for dead-beat control.
    """

    gain: float = 1.0

    def compute_task_dynamics(
        self, robot: pin.RobotWrapper
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the matrix :math:`J(q)` and vector :math:`\\alpha e(q)` such
        that the task dynamics are:

        .. math::

            J(q) \\Delta q = \\alpha e(q)

        The Jacobian matrix is :math:`J(q) \\in \\mathbb{R}^{k \\times n_v}`,
        with :math:`n_v` the dimension of the robot's tangent space and
        :math:`k` the dimension of the task. The error vector :math:`e(q) \\in
        \\mathbb{R}^k` is multiplied by the task gain :math:`\\alpha \\in [0,
        1]`. The gain is usually 1 for dead-beat control (*i.e.* converge as
        fast as possible), but it can also be lower for some extra low-pass
        filtering.

        Both :math:`J(q)` and :math:`(e)` depend on the configuration :math:`q`
        of the robot. The configuration displacement :math:`\\Delta q` is the
        output of inverse kinematics.

        Args:
            robot: Robot model and its current configuration :math:`q`.

        Returns:
            Tuple :math:`(J, e)` of Jacobian matrix and error vector, both
            expressed in the body frame.
        """
        raise NotImplementedError

    def compute_qp_objective(
        self, robot: pin.RobotWrapper
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the matrix-vector pair :math:`(H, c)` such that the
        contribution of the task to the QP objective of the IK is:

        .. math::

            \\| J \\Delta q - \\alpha e \\|_{W}^2 = \\frac{1}{2} \\Delta q^T H
            \\Delta q + c^T q

        The weight matrix :math:`W \\in \\mathbb{R}^{k \\times k}` weighs and
        normalizes task coordinates to the same unit. The unit of the overall
        contribution is [cost]^2. The configuration displacement :math:`\\Delta
        q` is the output of inverse kinematics (we divide it by :math:`\\Delta
        t` to get a commanded velocity).

        Args:
            robot: Robot model and its current configuration :math:`q`.

        Returns:
            Pair :math:`(H, c)` of Hessian matrix and linear vector of the QP
            objective.
        """
        raise NotImplementedError

    def __repr__(self):
        """
        Human-readable representation of the task.
        """
        return f"Task(gain={self.gain})"



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pink', 'pink.models', 'pink.tasks']

package_data = \
{'': ['*'],
 'pink.models': ['jvrc_description/*',
                 'jvrc_description/meshes/*',
                 'jvrc_description/meshes/back/*',
                 'jvrc_description/rsdf/*',
                 'jvrc_description/rsdf/jvrc1/*',
                 'jvrc_description/urdf/*']}

install_requires = \
['numpy >=1.20.0', 'pin >=2.6.3', 'qpsolvers >=1.8.0']

extras_require = \
{'doc': ['sphinx']}

setup(name='pin-pink',
      version='0.0.2',
      description='Python inverse kinematics for your robot model based on Pinocchio.',
      author=None,
      author_email='Stéphane Caron <stephane.caron@normalesup.org>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )
